@@ -3,8 +3,8 @@
  *
  *  Copyright (C) 1994  Mike Jagdis
  *
- * $Id: bsdsocket.c,v 1.4 1995/02/13 10:13:05 mike Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/bsdsocket.c,v $
+ * $Id: bsdsocket.c,v 1.5 1996/07/26 12:03:32 mike Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/bsdsocket.c,v $
  */
 
 #include <linux/config.h>
@@ -38,13 +38,13 @@ bsd_connect(struct pt_regs *regs)
 	 * the second is the address family. With Linux the address family
 	 * occupies both bytes.
 	 */
-	addr = (char *)get_fs_long(((unsigned long *)regs->esp) + 2);
+	addr = get_syscall_parameter (regs, 1);
 	if ((error = verify_area(VERIFY_READ, addr, 2)))
 		return error;
 	s = get_fs_word(addr);
 	put_fs_word(s>>8, addr);
 
-	error = SYS(socketcall)(SYS_CONNECT, ((unsigned long *)regs->esp) + 1);
+	error = SYS(socketcall)(SYS_CONNECT, get_syscall_parameter (regs, 0));
 
 	put_fs_word(s, addr);
 	return error;
