@@ -6,8 +6,8 @@
  *  Written by Drew Sullivan.
  *  Rewritten by Mike Jagdis.
  *
- * $Id: ioctl.c,v 1.54 1997/11/05 09:32:19 jaggy Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/ioctl.c,v $
+ * $Id: ioctl.c,v 1.55.2.2 1998/08/21 23:27:41 jaggy Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/ioctl.c,v $
  */
 #include <linux/config.h>
 
@@ -91,10 +91,13 @@ do_ioctl(struct pt_regs *regs, int fd, unsigned long ioctl_num, void *arg)
 		case 'm':
 			return sco_ioctl_tape(fd, ioctl_num, (unsigned long)arg);
 #endif
-#ifdef EMU_SVR4
 		case 't':
-			return ibcs_ioctl_tape(fd, ioctl_num, (unsigned long)arg);
+#ifdef EMU_SVR4
+			if (current->personality == PER_SVR4)
+				return ibcs_ioctl_tape(fd, ioctl_num, (unsigned long)arg);
+			else
 #endif
+			return bsd_ioctl_termios(fd, ioctl_num, arg);
 
 		case 'f':
 			return bsd_ioctl_file(fd, ioctl_num, arg);
@@ -857,7 +860,9 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 				return 1; /* at least 1... (FIXME) */
 			}
 #endif
-			error = SYS(ioctl)(fd, FIONREAD, arg);
+			error = SYS(ioctl)(fd, TIOCINQ, arg);
+			if (error == -EINVAL)
+				return 0;
 			if (error)
 				return error;
 			return ((unsigned long)get_fs_long(arg) == 0
@@ -880,7 +885,7 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 			memcpy_fromfs(&it, arg, sizeof(struct strioctl));
 #ifdef IBCS_TRACE
 			if (ibcs_trace & TRACE_STREAMS)
-				printk(KERN_ERR "iBCS: STREAMS I_STR ioctl(%d, 0x%08lx, 0x%08lx)\n",
+				printk(KERN_DEBUG "iBCS: STREAMS I_STR ioctl(%d, 0x%08lx, 0x%08lx)\n",
 					fd, (unsigned long)it.cmd, (unsigned long)it.data);
 #endif
 			switch (it.cmd >> 8) {
@@ -906,7 +911,7 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 				return error;
 #ifdef IBCS_TRACE
 			if (ibcs_trace & TRACE_STREAMS)
-				printk(KERN_ERR "iBCS: %d STREAMS I_PUSH %s\n",
+				printk(KERN_DEBUG "iBCS: %d STREAMS I_PUSH %s\n",
 					fd, tmp);
 #endif
 			putname(tmp);
@@ -916,7 +921,7 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 		case 003: /* I_POP */
 #ifdef IBCS_TRACE
 			if (ibcs_trace & TRACE_STREAMS)
-				printk(KERN_ERR "iBCS: %d STREAMS I_POP\n", fd);
+				printk(KERN_DEBUG "iBCS: %d STREAMS I_POP\n", fd);
 #endif
 			return 0;
 
@@ -931,7 +936,7 @@ static int ibcs_ioctl_stream(struct pt_regs *regs,
 				return error;
 #ifdef IBCS_TRACE
 			if (ibcs_trace & TRACE_STREAMS)
-				printk(KERN_ERR "iBCS: %d STREAMS I_FIND %s\n",
+				printk(KERN_DEBUG "iBCS: %d STREAMS I_FIND %s\n",
 					fd, tmp);
 #endif
 #ifdef EMU_XTI
