@@ -3,8 +3,8 @@
  *
  *  Copyright (C) 1994, 1996  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: socket.c,v 1.7 1996/03/29 17:33:33 mike Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/socket.c,v $
+ * $Id: socket.c,v 1.8 1997/11/19 19:32:03 jaggy Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/socket.c,v $
  */
 
 #include <linux/config.h>
@@ -47,6 +47,23 @@ ibcs_setsockopt(unsigned long *sp)
 	optname = (int)get_fs_long(((unsigned long *)sp)+2);
 
 	switch (level) {
+		case 0: /* IPPROTO_IP aka SOL_IP */
+			/* This is correct for the SCO family. Hopefully
+			 * it is correct for other SYSV...
+			 */
+			optname--;
+			if (optname == 0)
+				optname = 4;
+			if (optname > 4) {
+				optname += 24;
+				if (optname <= 33)
+					optname--;
+				if (optname < 32 || optname > 36)
+					return -EINVAL;
+			}
+			put_fs_long(optname, ((unsigned long *)sp)+2);
+			break;
+
 		case 0xffff:
 			put_fs_long(SOL_SOCKET, ((unsigned long *)sp)+1);
 			optname = map_value(sopt_map, optname, 0);
@@ -120,6 +137,23 @@ ibcs_getsockopt(unsigned long *sp)
 #endif
 
 	switch (level) {
+		case 0: /* IPPROTO_IP aka SOL_IP */
+			/* This is correct for the SCO family. Hopefully
+			 * it is correct for other SYSV...
+			 */
+			optname--;
+			if (optname == 0)
+				optname = 4;
+			if (optname > 4) {
+				optname += 24;
+				if (optname <= 33)
+					optname--;
+				if (optname < 32 || optname > 36)
+					return -EINVAL;
+			}
+			put_fs_long(optname, ((unsigned long *)sp)+2);
+			break;
+
 		case 0xffff:
 			put_fs_long(SOL_SOCKET, ((unsigned long *)sp)+1);
 			optname = map_value(sopt_map, optname, 0);
