@@ -3,8 +3,8 @@
  *
  *  Copyright 1994-1996  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: socksys.c,v 1.50 1997/05/10 14:45:54 jaggy Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/socksys.c,v $
+ * $Id: socksys.c,v 1.52 1997/11/12 22:30:09 jaggy Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/socksys.c,v $
  */
 
 #include <linux/config.h>
@@ -48,37 +48,27 @@
 #include <ibcs/tli.h>
 #endif
 
+
+int socksys_major;
+
+
+static int socksys_read(struct inode *ino, struct file *filep,
+			char *buf, int count);
+static int socksys_write(struct inode *ino, struct file *filep,
+			const char *buf, int count);
 static int socksys_open(struct inode *ino, struct file *filep);
 static void socksys_close(struct inode *ino, struct file *filep);
 static int socksys_select(struct inode *ino, struct file *filep,
 			int sel_type, select_table *wait);
 
-static int spx_write(struct inode *ino, struct file *filep,
-			const char *buf, int count);
-
-/* spx_fops defines the file operations that can be applied to the
- * /dev/spx server devices.
- */
-static struct file_operations spx_fops = {
-	NULL,			/* lseek */
-	NULL,			/* read */
-	spx_write,		/* write */
-	NULL,			/* readdir */
-	NULL,			/* select */
-	NULL,			/* ioctl - ibcs_ioctl calls ibcs_ioctl_socksys*/
-	NULL,			/* mmap */
-	NULL,			/* open */
-	socksys_close,		/* close */
-	NULL			/* fsync */
-};
 
 /* socksys_fops defines the file operations that can be applied to the
  * /dev/socksys device.
  */
 static struct file_operations socksys_fops = {
 	NULL,			/* lseek */
-	NULL,			/* read */
-	NULL,			/* write */
+	socksys_read,		/* read */
+	socksys_write,		/* write */
 	NULL,			/* readdir */
 	NULL,			/* select */
 	NULL,			/* ioctl - ibcs_ioctl calls ibcs_ioctl_socksys*/
@@ -134,8 +124,286 @@ inherit_socksys_funcs(struct file *filep, int fd)
 }
 
 
-static inline int
-socksys_syscall(int fd, int *sp)
+static int
+spx_connect(struct inode *ino, struct file *filep, int spxnum)
+{
+	int newfd, err, args[3];
+	struct sockaddr_un addr = {
+		AF_UNIX, "/tmp/.X11-unix/X0"
+	};
+	int old_fs = get_fs();
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] %lx choose service %d\n",
+			current->pid, (unsigned long)filep, spxnum);
+#endif
+
+	/* Rather than use an explicit path to the X :0 server
+	 * socket we should use the given number to look up a path
+	 * name to use (we can't rely on servers registering their
+	 * sockets either - for one thing we don't emulate that yet
+	 * and for another thing different OS binaries do things in
+	 * different ways but all must interoperate).
+	 * I suggest putting the mapping in, say, /dev/spx.map/%d
+	 * where each file is a symlink containing the path of the
+	 * socket to use. Then we can just do a readlink() here to
+	 * get the pathname.
+	 *   Hey, this is what we do here now!
+	 */
+	addr.sun_family = AF_UNIX;
+	sprintf(addr.sun_path, "/dev/spx.map/%d", spxnum);
+	set_fs(get_ds());
+	err = SYS(readlink)(addr.sun_path, addr.sun_path,
+				sizeof(addr.sun_path)-1);
+	set_fs(old_fs);
+	if (err == -ENOENT) {
+#ifdef IBCS_TRACE
+		if (ibcs_trace & TRACE_SOCKSYS)
+			printk(KERN_DEBUG
+				"SPX: [%d] %lx no symlink \"%s\", try X :0\n",
+				current->pid, (unsigned long)filep,
+				addr.sun_path);
+#endif
+		strcpy(addr.sun_path, "/tmp/.X11-unix/X0");
+	} else {
+		if (err < 0)
+			return err;
+		addr.sun_path[err] = '\0';
+	}
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] %lx get a Unix domain socket\n",
+			current->pid, (unsigned long)filep);
+#endif
+	args[0] = AF_UNIX;
+	args[1] = SOCK_STREAM;
+	args[2] = 0;
+	set_fs(get_ds());
+	newfd = SYS(socketcall)(SYS_SOCKET, args);
+	set_fs(old_fs);
+	if (newfd < 0)
+		return newfd;
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "SPX: [%d] %lx connect to \"%s\"\n",
+			current->pid, (unsigned long)filep, addr.sun_path);
+#endif
+	args[0] = newfd;
+	args[1] = (int)&addr;
+	args[2] = sizeof(struct sockaddr_un);
+	set_fs(get_ds());
+	err = SYS(socketcall)(SYS_CONNECT, args);
+	set_fs(old_fs);
+	if (err) {
+		SYS(close)(newfd);
+		return err;
+	}
+	return newfd;
+}
+
+
+int
+ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count)
+{
+	struct file *filep = current->FD[fd];
+	struct inode *ino = current->FD[fd]->f_inode;
+	int error, sockfd;
+
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG
+			"socksys: [%d] fd=%d initializing\n",
+			current->pid, fd);
+#endif
+	/* Minor = 0 is the socksys device itself. No special handling
+	 *           will be needed as it is controlled by the application
+	 *           via ioctls.
+	 */
+	if (MINOR(ino->i_rdev) == 0)
+		return 0;
+
+	/* Minor = 1 is the spx device. This is the client side of a
+	 *           streams pipe to the X server. Under SCO and friends
+	 *           the library code messes around setting the connection
+	 *           up itself. We do it ourselves - this means we don't
+	 *           need to worry about the implementation of the server
+	 *           side (/dev/X0R - which must exist but can be a link
+	 *           to /dev/null) nor do we need to actually implement
+	 *           getmsg/putmsg.
+	 */
+	if (MINOR(ino->i_rdev) == 1) {
+		int unit = 1;
+
+		/* It seems early spx implementations were just a
+		 * quick hack to get X to work. They only supported
+		 * one destination and connected automatically.
+		 * Later versions take a single byte write, the
+		 * value of the byte telling them which destination
+		 * to connect to.
+		 */
+		if (rw == 1 && count == 1) {
+			error = verify_area(VERIFY_READ, buf, 1*sizeof(char));
+			if (error)
+				return error;
+
+			unit = get_fs_byte(buf);
+		}
+
+		sockfd = spx_connect(ino, filep, unit);
+		if (sockfd < 0)
+			return sockfd;
+	}
+
+	/*
+	 * Otherwise the high 4 bits specify the address/protocol
+	 * family (AF_INET, AF_UNIX etc.) and the low 4 bits determine
+	 * the protocol (IPPROTO_IP, IPPROTO_UDP, IPPROTO_TCP etc.)
+	 * although not using a one-to-one mapping as the minor number
+	 * is not big enough to hold everything directly. The socket
+	 * type is inferrred from the protocol.
+	 */
+	else { /* XTI */
+#ifdef EMU_XTI
+		struct T_private *priv;
+#endif
+		int args[3];
+		int old_fs = get_fs();
+
+		/* Grab a socket. */
+#ifdef IBCS_TRACE
+		if (ibcs_trace & TRACE_SOCKSYS)
+			printk(KERN_DEBUG
+				"XTI: [%d] %d get socket for transport end"
+				" point (dev = 0x%04x)\n",
+				current->pid, fd, ino->i_rdev);
+#endif
+		switch ((args[0] = ((MINOR(ino->i_rdev) >> 4) & 0x0f))) {
+			case AF_UNIX:
+				args[1] = SOCK_STREAM;
+				args[2] = 0;
+				break;
+
+			case AF_INET: {
+				int prot[16] = {
+					IPPROTO_ICMP,	IPPROTO_ICMP,
+					IPPROTO_IGMP,	IPPROTO_IPIP,
+					IPPROTO_TCP,	IPPROTO_EGP,
+					IPPROTO_PUP,	IPPROTO_UDP,
+					IPPROTO_IDP,	IPPROTO_RAW,
+				};
+				int type[16] = {
+					SOCK_RAW,	SOCK_RAW,
+					SOCK_RAW,	SOCK_RAW,
+					SOCK_STREAM,	SOCK_RAW,
+					SOCK_RAW,	SOCK_DGRAM,
+					SOCK_RAW,	SOCK_RAW,
+				};
+				int i = MINOR(ino->i_rdev) & 0x0f;
+				args[2] = prot[i];
+				args[1] = type[i];
+				break;
+			}
+
+			default:
+				args[1] = SOCK_RAW;
+				args[2] = 0;
+				break;
+		}
+#ifdef IBCS_TRACE
+		if (ibcs_trace & TRACE_SOCKSYS)
+			printk(KERN_DEBUG
+				"XTI: [%d] %d socket %d %d %d\n",
+				current->pid, fd,
+				args[0], args[1], args[2]);
+#endif
+		set_fs(get_ds());
+		sockfd = SYS(socketcall)(SYS_SOCKET, args);
+		set_fs(old_fs);
+		if (sockfd < 0)
+			return sockfd;
+
+#ifdef EMU_XTI
+		priv = (struct T_private *)kmalloc(sizeof(struct T_private), GFP_KERNEL);
+		if (!priv) {
+			SYS(close)(sockfd);
+			return -ENOMEM;
+		}
+		priv->state = TS_UNBND;
+		priv->offset = 0;
+		priv->pfirst = priv->plast = NULL;
+		current->FD[sockfd]->private_data = priv;
+#endif
+	}
+
+	/* Redirect operations on the socket fd via our emulation
+	 * handlers then swap the socket fd and the original fd,
+	 * discarding the original fd.
+	 */
+	inherit_socksys_funcs(current->FD[sockfd], sockfd);
+#ifdef IBCS_TRACE
+	if (ibcs_trace & TRACE_SOCKSYS)
+		printk(KERN_DEBUG "XTI: [%d] %d (ino=%lx) -> %d (ino=%lx)\n",
+			current->pid,
+			fd, (unsigned long)current->FD[fd]->f_inode,
+			sockfd, (unsigned long)current->FD[sockfd]->f_inode);
+#endif
+#if 0
+	current->FD[sockfd]->f_inode->i_rdev = ino->i_rdev;
+#endif
+	MOD_INC_USE_COUNT;
+	SYS(dup2)(sockfd, fd);
+	SYS(close)(sockfd);
+	return 0;
+}
+
+
+static int
+socksys_read(struct inode *ino, struct file *filep, char *buf, int count)
+{
+	if (ino && !ino->i_sock) {
+		int fd;
+		for (fd=0; fd<NR_OPEN; fd++) {
+			if (current->FD[fd] == filep) {
+				int error;
+				error = ibcs_socksys_fd_init(fd, 0, NULL, 0);
+				if (error)
+					return error;
+				filep = current->FD[fd];
+				ino = filep->f_inode;
+				break;
+			}
+		}
+	}
+	return filep->f_op->read(ino, filep, buf, count);
+}
+
+
+static int
+socksys_write(struct inode *ino, struct file *filep, const char *buf, int count)
+{
+	if (ino && !ino->i_sock) {
+		int fd;
+		for (fd=0; fd<NR_OPEN; fd++) {
+			if (current->FD[fd] == filep) {
+				int error;
+				error = ibcs_socksys_fd_init(fd, 1, buf, count);
+				if (error)
+					return error;
+				filep = current->FD[fd];
+				ino = filep->f_inode;
+				break;
+			}
+		}
+	}
+	return filep->f_op->write(ino, filep, buf, count);
+}
+
+
+int
+socksys_syscall(int *sp)
 {
 	int cmd;
 
@@ -161,8 +429,8 @@ socksys_syscall(int fd, int *sp)
 		a3 = (unsigned long)get_fs_long(sp+3);
 		a4 = (unsigned long)get_fs_long(sp+4);
 		a5 = (unsigned long)get_fs_long(sp+5);
-		printk(KERN_DEBUG "socksys: [%d] %d: %s (%d) <0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx>\n",
-			current->pid, fd,
+		printk(KERN_DEBUG "socksys: [%d]: %s (%d) <0x%lx,0x%lx,0x%lx,0x%lx,0x%lx,0x%lx>\n",
+			current->pid,
 			(cmd >= 0 && cmd < sizeof(cmd_map)/sizeof(cmd_map[0]))
 				? cmd_map[cmd] : "???",
 			cmd, a0, a1, a2, a3, a4, a5);
@@ -535,7 +803,7 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 
 		case SSYS_SIOCSOCKSYS:		/* Pseudo socket syscall */
 		case SVR4_SIOCSOCKSYS:
-			return socksys_syscall(fd, (int *)arg);
+			return socksys_syscall((int *)arg);
 
 		case SSYS_SIOCSHIWAT:		/* set high watermark */
 		case SVR4_SIOCSHIWAT:
@@ -728,118 +996,6 @@ ibcs_ioctl_socksys(int fd, unsigned int cmd, void *arg)
 }
 
 
-static int
-spx_connect(struct inode *ino, struct file *filep, int spxnum)
-{
-	int newfd, err, args[3];
-	struct sockaddr_un addr = {
-		AF_UNIX, "/tmp/.X11-unix/X0"
-	};
-	int old_fs = get_fs();
-
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] %lx choose service %d\n",
-			current->pid, (unsigned long)filep, spxnum);
-#endif
-
-	/* Rather than use an explicit path to the X :0 server
-	 * socket we should use the given number to look up a path
-	 * name to use (we can't rely on servers registering their
-	 * sockets either - for one thing we don't emulate that yet
-	 * and for another thing different OS binaries do things in
-	 * different ways but all must interoperate).
-	 * I suggest putting the mapping in, say, /dev/spx.map/%d
-	 * where each file is a symlink containing the path of the
-	 * socket to use. Then we can just do a readlink() here to
-	 * get the pathname.
-	 *   Hey, this is what we do here now!
-	 */
-	addr.sun_family = AF_UNIX;
-	sprintf(addr.sun_path, "/dev/spx.map/%d", spxnum);
-	set_fs(get_ds());
-	err = SYS(readlink)(addr.sun_path, addr.sun_path,
-				sizeof(addr.sun_path)-1);
-	set_fs(old_fs);
-	if (err == -ENOENT) {
-#ifdef IBCS_TRACE
-		if (ibcs_trace & TRACE_SOCKSYS)
-			printk(KERN_DEBUG
-				"SPX: [%d] %lx no symlink \"%s\", try X :0\n",
-				current->pid, (unsigned long)filep,
-				addr.sun_path);
-#endif
-		strcpy(addr.sun_path, "/tmp/.X11-unix/X0");
-	} else {
-		if (err < 0)
-			return err;
-		addr.sun_path[err] = '\0';
-	}
-
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] %lx get a Unix domain socket\n",
-			current->pid, (unsigned long)filep);
-#endif
-	args[0] = AF_UNIX;
-	args[1] = SOCK_STREAM;
-	args[2] = 0;
-	set_fs(get_ds());
-	newfd = SYS(socketcall)(SYS_SOCKET, args);
-	set_fs(old_fs);
-	if (newfd < 0)
-		return newfd;
-
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] %lx connect to \"%s\"\n",
-			current->pid, (unsigned long)filep, addr.sun_path);
-#endif
-	args[0] = newfd;
-	args[1] = (int)&addr;
-	args[2] = sizeof(struct sockaddr_un);
-	set_fs(get_ds());
-	err = SYS(socketcall)(SYS_CONNECT, args);
-	set_fs(old_fs);
-	if (err) {
-		SYS(close)(newfd);
-		return err;
-	}
-
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "SPX: [%d] filep=0x%08lx, swap inodes old=0x%08lx, new=0x%08lx\n",
-			current->pid, (unsigned long)filep, (unsigned long)ino,
-			(unsigned long)current->FD[newfd]->f_inode);
-#endif
-
-	inherit_socksys_funcs(filep, newfd);
-	filep->f_inode = current->FD[newfd]->f_inode;
-	filep->f_inode->i_rdev = ino->i_rdev;
-	filep->f_inode->i_flock = ino->i_flock;
-	filep->f_inode->u.socket_i.file = filep;
-	iput(ino);
-	FD_CLR(newfd, &current->files->close_on_exec);
-	FD_CLR(newfd, &current->files->open_fds);
-	current->FD[newfd]->f_count--;
-	current->FD[newfd] = NULL;
-
-	return 1;
-}
-
-
-static int
-spx_write(struct inode *ino, struct file *filep, const char *buf, int count)
-{
-	int error;
-
-	if ((error = verify_area(VERIFY_READ, buf, 1*sizeof(char))))
-		return error;
-
-	return spx_connect(ino, filep, get_fs_byte(buf));
-}
-
-
 static int
 socksys_open(struct inode *ino, struct file *filep)
 {
@@ -850,146 +1006,7 @@ socksys_open(struct inode *ino, struct file *filep)
 			"socksys: [%d] filep=0x%08lx, inode=0x%08lx opening\n",
 			current->pid, (unsigned long)filep, (unsigned long)ino);
 #endif
-	/* Minor = 0 is the socksys device itself. No special handling
-	 *           will be needed as it is controlled by the application
-	 *           via ioctls.
-	 */
-	if (MINOR(ino->i_rdev) == 0)
-		return 0;
-
-	/* Minor = 1 is the spx device. This is the client side of a
-	 *           streams pipe to the X server. Under SCO and friends
-	 *           the library code messes around setting the connection
-	 *           up itself. We do it ourselves - this means we don't
-	 *           need to worry about the implementation of the server
-	 *           side (/dev/X0R - which must exist but can be a link
-	 *           to /dev/null) nor do we need to actually implement
-	 *           getmsg/putmsg.
-	 */
-	if (MINOR(ino->i_rdev) == 1) {
-		filep->f_op = &spx_fops;
-#if 0
-		/* It seems that there was an early spx implementation
-		 * that only supported one destination and which
-		 * required nothing but an open() to connect. When we
-		 * find out how to identify such cases we can wrap this
-		 * code in a conditional and enable it. In the meantime
-		 * you might find "ln -s /tmp/.X11-unix/X0 /dev/spx"
-		 * will work?
-		 */
-		return spx_connect(ino, filep, 1);
-#else
-		return 0;
-#endif
-	}
-
-	/*
-	 * Otherwise the high 4 bits specify the address/protocol
-	 * family (AF_INET, AF_UNIX etc.) and the low 4 bits determine
-	 * the protocol (IPPROTO_IP, IPPROTO_UDP, IPPROTO_TCP etc.)
-	 * although not using a one-to-one mapping as the minor number
-	 * is not big enough to hold everything directly. The socket
-	 * type is inferrred from the protocol.
-	 */
-{ /* XTI */
-#ifdef EMU_XTI
-	struct T_private *priv;
-#endif
-	int fd, args[3];
-	int old_fs = get_fs();
-
-	/* Grab a socket. */
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG
-			"XTI: [%d] %lx get socket for transport end point"
-			" (dev = 0x%04x)\n",
-			current->pid, (unsigned long)filep,
-			ino->i_rdev);
-#endif
-	switch ((args[0] = ((MINOR(ino->i_rdev) >> 4) & 0x0f))) {
-		case AF_UNIX:
-			args[0] = AF_UNIX;
-			args[1] = SOCK_STREAM;
-			args[2] = 0;
-			break;
-
-		case AF_INET: {
-			int prot[16] = {
-				IPPROTO_ICMP,	IPPROTO_ICMP,	IPPROTO_IGMP,
-				IPPROTO_IPIP,	IPPROTO_TCP,	IPPROTO_EGP,
-				IPPROTO_PUP,	IPPROTO_UDP,	IPPROTO_IDP,
-				IPPROTO_RAW,
-			};
-			int type[16] = {
-				SOCK_RAW,	SOCK_RAW,	SOCK_RAW,
-				SOCK_RAW,	SOCK_STREAM,	SOCK_RAW,
-				SOCK_RAW,	SOCK_DGRAM,	SOCK_RAW,
-				SOCK_RAW,
-			};
-			int i = MINOR(ino->i_rdev) & 0x0f;
-			args[2] = prot[i];
-			args[1] = type[i];
-			break;
-		}
-
-		default:
-			args[1] = SOCK_RAW;
-			args[2] = 0;
-			break;
-	}
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG
-			"XTI: [%d] %lx socket %d %d %d\n",
-			current->pid, (unsigned long)filep,
-			args[0], args[1], args[2]);
-#endif
-	set_fs(get_ds());
-	fd = SYS(socketcall)(SYS_SOCKET, args);
-	set_fs(old_fs);
-	if (fd < 0) {
-		MOD_DEC_USE_COUNT;
-		return fd;
-	}
-
-#ifdef EMU_XTI
-	priv = (struct T_private *)kmalloc(sizeof(struct T_private), GFP_KERNEL);
-	if (!priv) {
-		SYS(close)(fd);
-		MOD_DEC_USE_COUNT;
-		return -ENOMEM;
-	}
-	priv->state = TS_UNBND;
-	priv->offset = 0;
-	priv->pfirst = priv->plast = NULL;
-#endif
-
-	/* Release the inode we were given, replace with the socket
-	 * inode, redirect operations to our emulation handlers then
-	 * clear down the descriptor used for the socket.
-	 */
-#ifdef IBCS_TRACE
-	if (ibcs_trace & TRACE_SOCKSYS)
-		printk(KERN_DEBUG "XTI: [%d] %lx swap inodes\n",
-			current->pid, (unsigned long)filep);
-#endif
-	inherit_socksys_funcs(filep, fd);
-	filep->f_inode = current->FD[fd]->f_inode;
-	filep->f_inode->i_rdev = ino->i_rdev;
-	filep->f_inode->i_flock = ino->i_flock;
-	filep->f_inode->u.socket_i.file = filep;
-	iput(ino);
-#ifdef EMU_XTI
-	filep->private_data = priv;
-#endif
-	FD_CLR(fd, &current->files->close_on_exec);
-	FD_CLR(fd, &current->files->open_fds);
-	current->FD[fd]->f_count--;
-	current->FD[fd] = NULL;
-
 	return 0;
-} /* XTI */
 }
 
 
@@ -1020,7 +1037,7 @@ socksys_close(struct inode *ino, struct file *filep)
 {
 	if (ino && ino->i_sock) {
 #ifdef EMU_XTI
-		if (MINOR(ino->i_rdev) != 1 && filep->private_data) {
+		if (filep->private_data) {
 			struct T_primsg *it;
 			it = ((struct T_private *)filep->private_data)->pfirst;
 			while (it) {
@@ -1043,9 +1060,6 @@ socksys_close(struct inode *ino, struct file *filep)
 }
 
 
-static int socksys_major;
-
-
 void
 init_socksys(void)
 {
