@@ -3,8 +3,8 @@
  *
  *  Copyright (C) 1994  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: sysconf.c,v 1.7 1997/01/05 18:19:10 jaggy Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/sysconf.c,v $
+ * $Id: sysconf.c,v 1.8 1997/07/11 21:28:18 jaggy Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/sysconf.c,v $
  */
 
 #include <linux/config.h>
@@ -63,51 +63,97 @@
 
 #define _SC_PAGESIZE   11
 
-int ibcs_sysconf(int name) {
+
+/* This is an SVR4 system call that is undocumented except for some
+ * hints in a header file. It appears to be a forerunner to the
+ * POSIX sysconf() call.
+ */
+int ibcs_sysconfig(int name)
+{
 	switch (name) {
-		case _SC_ARG_MAX: {
-			/* From limits.h */
-			return (ARG_MAX);
-		}
-
-		case _SC_CHILD_MAX: {
-			/* From limits.h */
-			return (CHILD_MAX);
-		}
-
-		case _SC_CLK_TCK: {
-			return (HZ);
-		}
-
-		case _SC_NGROUPS_MAX: {
+		case 2: /* _CONFIG_NGROUPS */
 			/* From limits.h */
 			return (NGROUPS_MAX);
-		}
 
-		case _SC_OPEN_MAX: {
+		case 3: /* _CONFIG_CHILD_MAX */
+			/* From limits.h */
+			return (CHILD_MAX);
+
+		case 4: /* _CONFIG_OPEN_FILES */
 			/* From limits.h */
 			return (OPEN_MAX);
-		}
 
-		case _SC_JOB_CONTROL: {
+		case 5: /* _CONFIG_POSIX_VER */
+			/* The version of the POSIX standard we conform
+			 * to. SCO defines _POSIX_VERSION as 198808L
+			 * sys/unistd.h. What are we?
+			 * (N.B. SCO does not even do sysconfig but I
+			 * have no SVR4 running)
+			 */
+			return (198808L);
+
+		case 6: /* _CONFIG_PAGESIZE */
+			return PAGE_SIZE;
+
+		case 7: /* _CONFIG_CLK_TCK */
+			return (HZ);
+
+		case 8: /* _CONFIG_XOPEN_VER */
+			return -EINVAL; /* FIXME: huh? */
+
+		case 9: /* _CONFIG_NACLS_MAX */
+			return -EINVAL; /* FIXME: huh? */
+
+		case 10: /* _CONFIG_ARG_MAX */
+			/* From limits.h */
+			return (ARG_MAX);
+	}
+#ifdef IBCS_TRACE
+	if ((ibcs_trace & TRACE_API) || ibcs_func_p->trace) {
+		printk(KERN_DEBUG "iBCS2 unsupported sysconfig call %d\n", name);
+	}
+#endif
+
+	return -EINVAL;
+}
+
+
+int ibcs_sysconf(int name) {
+	switch (name) {
+		case _SC_ARG_MAX:
+			/* From limits.h */
+			return (ARG_MAX);
+
+		case _SC_CHILD_MAX:
+			/* From limits.h */
+			return (CHILD_MAX);
+
+		case _SC_CLK_TCK:
+			return (HZ);
+
+		case _SC_NGROUPS_MAX:
+			/* From limits.h */
+			return (NGROUPS_MAX);
+
+		case _SC_OPEN_MAX:
+			/* From limits.h */
+			return (OPEN_MAX);
+
+		case _SC_JOB_CONTROL:
 			return (1);
-		}
 
-		case _SC_SAVED_IDS: {
+		case _SC_SAVED_IDS:
 			return (1);
-		}
 
-		case _SC_PAGESIZE: {
-		  return PAGE_SIZE;
-		}
+		case _SC_PAGESIZE:
+			return PAGE_SIZE;
 
-		case _SC_VERSION: {
+		case _SC_VERSION:
 			/* The version of the POSIX standard we conform
 			 * to. SCO defines _POSIX_VERSION as 198808L
 			 * sys/unistd.h. What are we?
 			 */
 			return (198808L);
-		}
 	}
 
 #ifdef IBCS_TRACE
