@@ -6,8 +6,8 @@
  *  Hacked by Eric Youngdale for iBCS (1993, 1994).
  *  Added to by Drew Sullivan, modified by EY for xstat (used by SVr4).
  *
- * $Id: xstat.c,v 1.12 1997/05/13 19:56:30 jaggy Exp $
- * $Source: /usr/CVS/ibcs/iBCSemul/xstat.c,v $
+ * $Id: xstat.c,v 1.13 1998/02/02 23:02:52 jaggy Exp $
+ * $Source: /u/CVS/ibcs/iBCSemul/xstat.c,v $
  */
 
 #include <linux/config.h>
@@ -21,9 +21,6 @@
 #endif
 
 #include <linux/kernel.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <sys/time.h>
 #include <linux/fs.h>
 #include <linux/errno.h>
 #include <linux/mm.h>
