@@ -1,8 +1,8 @@
 /*
  *  Function prototypes used by the iBCS2 emulator
  *
- * $Id: ibcs.h,v 1.48 1997/05/10 14:45:55 jaggy Exp $
- * $Source: /usr/CVS/ibcs/include/ibcs/ibcs.h,v $
+ * $Id: ibcs.h,v 1.53 1997/11/12 22:30:10 jaggy Exp $
+ * $Source: /u/CVS/ibcs/include/ibcs/ibcs.h,v $
  */
 #include <linux/ptrace.h>	/* for pt_regs */
 #include <linux/sched.h>
@@ -164,7 +164,7 @@ extern void iABI_emulate(struct pt_regs *regs);
 #if !defined(CONFIG_BINFMT_IBCS) && !defined(INIT_MM)
 extern void ibcs_exit(int n);
 #endif
-extern int iABI_errors(int errno);
+extern int iABI_errors(int lnx_errno);
 extern int ibcs_syscall(struct pt_regs *regs);
 
 /* fcntl.c */
@@ -282,13 +282,11 @@ extern int svr4_access(char *path, int mode);
 extern int ibcs_sysinfo(int, char *, long);
 
 /* sysconf.c */
+extern int ibcs_sysconfig(int name);
 extern int ibcs_sysconf(int name);
 
-/* Kernels 1.1.10 and later have sysfs as a system call. */
-#ifndef __NR_sysfs
 /* sysfs.c */
-extern int ibcs_sysfs(struct pt_regs * regs);
-#endif
+extern int ibcs_sysfs(int cmd, int arg1, int arg2);
 
 /* sysi86.c */
 extern int ibcs_sysi86(struct pt_regs * regs);
@@ -330,6 +328,9 @@ extern int wv386_getsockname(struct pt_regs *regs);
 extern int wv386_ioctl(int fd, unsigned int ioctl_num, void *arg);
 
 /* From socksys.c */
+extern int socksys_major;
+extern int ibcs_socksys_fd_init(int fd, int rw, const char *buf, int count);
+extern int socksys_syscall(int *sp);
 extern void init_socksys(void);
 extern void cleanup_socksys(void);
 
