@@ -3,14 +3,17 @@
  *
  *  Copyright (C) 1994  Mike Jagdis (jaggy@purplet.demon.co.uk)
  *
- * $Id: stream.h,v 1.2 1995/07/03 12:52:12 mike Exp $
- * $Source: /u3/CVS/ibcs/include/ibcs/stream.h,v $
+ * $Id: stream.h,v 1.4 1996/01/24 10:22:42 mike Exp $
+ * $Source: /u/CVS/ibcs/include/ibcs/stream.h,v $
  */
 
 #ifndef _IBCS_STREAM_H_
 #define _IBCS_STREAM_H_
 
-#define RS_HIPRI	1
+#define MSG_HIPRI	1
+#define RS_HIPRI	MSG_HIPRI
+#define MSG_ANY		2
+#define MSG_BAND	4
 
 #define MORECTL		1
 #define MOREDATA	2
@@ -21,4 +24,11 @@ struct strbuf {
 	char	*buf;		/* pointer to buffer */
 };
 
+/* Used for the I_PEEK STREAMS ioctl. */
+struct strpeek {
+	struct strbuf ctl;
+	struct strbuf dat;
+	long flags;
+};
+
 #endif
