@@ -4,8 +4,8 @@
  * Copyright (C) 1993  Drew Sullivan
  * Released for general use as long as this copyright remains.
  *
- * $Id: xnx.h,v 1.3 1994/05/27 10:49:54 mike Exp $
- * $Source: /u3/CVS/ibcs/include/ibcs/xnx.h,v $
+ * $Id: xnx.h,v 1.4 1996/06/06 16:23:05 mike Exp $
+ * $Source: /u/CVS/ibcs/include/ibcs/xnx.h,v $
  */
 typedef unsigned short excode_t;
 
@@ -16,7 +16,7 @@ struct timeb {
 	short dstflag;
 };
 
-extern	int xnx_locking(int fd, int mode, long size);
+extern	int xnx_locking(int fd, int mode, unsigned long size);
 extern	int xnx_creatsem(char *sem_name, int mode);
 extern	int xnx_opensem(char *sem_name);
 extern	int xnx_sigsem(int sem_num);
